//#include "tnt_array1d.h"

#include "LevyProc.h"

#define SQR(x) ((x)*(x))

extern "C"
{
  double S_STDCALL S_rpois( double lambda );
  double S_STDCALL S_runif( void );
  double S_STDCALL S_rexp( void );
  double S_STDCALL S_expm1( double x );
}

/**************************************************************************
 **  e1 and e1inv
 **  
 **  Exponential Integral Function and its inverse (see A+S sec 5.1) 
 **  Note this is only used to draw the initial MCMC draw, so it
 **  is not critical.... the qchisq/pchisq approx'ns are good enough.
 **  The following work well for small x, large y... say, for x < 1 < y.
 **
 **/
double e1( double x ) {      // from A&S 5.1.53:
  return( -log(x) - 0.57721566 + x*(0.99999193 - x*(0.24991055 - 
    x*(0.55199680 - x*(0.00976004 - x*(0.00107857))))) );
}
double e1inv( double y ) {   // AIJM van Dijk (2002)  
  double t1 = exp(-y-0.577215665),
    t2 = exp(-2.216*y-0.722),
    t3 = exp(-6.547*y-2.481);
  return(t1+t2-t3);
}



/**************************************************************************
 **  Constructor
 **
 **/
LevyProc::LevyProc(
  double *xlim, double *ylim, int JmpMax, double agam, 
  double rgam, double epsilon, double dist
  )
{
  JmpMax_ = JmpMax;
  agam_ = agam;
  rgam_ = rgam;
  dist_ = dist;
  epsilon_ = epsilon;        
  // If region not specified, default to unit square.
  if( (xlim == NULL) || (ylim == NULL) ){         
    xlim_ = new double (2);
    ylim_ = new double (2);
    xlim_[0] = 0;
    xlim_[1] = 1;
    ylim_[0] = 0;
    ylim_[1] = 1;
  } else {
    xlim_ = xlim;
    ylim_ = ylim;
  }
  area = ( xlim_[1] - xlim_[0] ) * (ylim_[1] - ylim_[0] );
  #ifdef DEBUG1  
    printf("area = %f\n", area);
  #endif
  atomHeight_ = vector<double> ();
  xLoc_ = vector<double> ();
  yLoc_ = vector<double> ();

  nAtoms_ = 0;
}

/**************************************************************************
 **  Destructor
 **
 **/
LevyProc::~LevyProc(){}

/**************************************************************************
 **  Draw a location uniformly from the rectangular region.  Put the coords 
 **  in x and y.
 **
 **/
void LevyProc::drawUnifLoc(double *x, double *y){

  (*x) = xlim_[0] + (xlim_[1] - xlim_[0]) * S_runif();
  (*y) = ylim_[0] + (ylim_[1] - ylim_[0]) * S_runif();
}

/**************************************************************************
 **  Draw from the Gamma random field prior: generate random draw of size M 
 **  within the rectangular region.
 **
 **/
void LevyProc::drawFromPrior() {
  
  double EM = agam_ * area * e1( rgam_*epsilon_ );    // Exp num E[M] jmps > eps
  #ifdef DEBUG1  
    printf( "EM = %f; agam = %f; rgam = %f; area = %f\n", EM, agam_, rgam_, area);
  #endif
  if( (JmpMax_ > 0) && (EM > JmpMax_) ) {
    printf( "Reducing EM=%f to JmpMax=%d\n", EM, JmpMax_ );
    EM = JmpMax_;
  }
  int nAtomsToAdd = (int) S_rpois( EM );
  nAtomsToAdd = std::max( nAtomsToAdd, 1 );

  double x, y, z;
  for (int i=0; i<nAtomsToAdd; i++){
    drawUnifLoc ( &x, &y );
    z = S_runif() * EM;      // Jump times, cont'l on M
    z = e1inv( z/( agam_*area ) ) / rgam_;    // Levy proc mass sizes
    z = z + epsilon_;  // ensure that z is larger than epsilon
    addPoint( x, y, z ); 
  }
  #ifdef DEBUG1
    printf( "Natoms = %d\n", nAtoms_);
  #endif
}

/**************************************************************************
 **  Calculate the squared distance of z from the height of the Levy
 **  process surface at the location (x,y), on the log scale.
 **  datGeo indicates whether the distance should take into account
 **  the curvature of the earth.
 **
 **
double LevyProc::squaredError( double x, 
  double y, double z, 
  bool datGeo ){

  double costh;
  const double R = 6378.137,      // Earth's radius in km.
    deg2rad = 0.01745329;         // pi/180
  if( dist_ == 0.0 ){
    MESSAGE "Dist_ = 0\n" ERROR;
  }

  vector<double>::iterator itX,
    itY = yLoc_.begin(),
    itHeight = atomHeight_.begin();

  double YHat = 0.0,
    d2, startTime;

  // The slow part of the following calculation may be the 
  // exponential function
  for( itX = xLoc_.begin(); itX < xLoc_.end(); itX++ ){
      
    d2 = SQR((*itX) - x) + SQR((*itY) - y);
    //    startTime = tk.clockSecs();
    YHat += (*itHeight) * exp(-0.5 * d2 / SQR(dist_) );
    //    expCompTime += tk.clockSecs() - startTime;
    itY++;  
    itHeight++;
  }


  // Alternate slower code
  for(int j=0; j< nAtoms_; j++) {
    if(! datGeo) { d2 = SQR(x - xLoc_[j]) + SQR(y - yLoc_[j]);
    } else {                   // Spherical geometry
      costh = 
	cos(y*deg2rad) * cos(yLoc_[j]*deg2rad) * cos((x-xLoc_[j])*deg2rad) +
	sin(y*deg2rad) * sin(yLoc_[j]*deg2rad);
      d2 = SQR(R * acos(costh));
    }
    YHat += atomHeight_[j] * exp(-0.5 * d2 / SQR(dist_) );
    }

  if( ( z <= 0.0 ) || ( YHat <= 0.0 ) ){
    MESSAGE "z<=0 or Yhat <= 0\n" ERROR;
  }
  return( SQR( log( z / YHat ) ) );
}*/

/**************************************************************************
 **  Calculate the squared distance of z from the height of the Levy
 **  process surface at the location (x,y), on the log scale.
 **  datGeo indicates whether the distance should take into account
 **  the curvature of the earth.
 **
 **/
double LevyProc::sumSquaredError( const Array1D<double> &x, 
  const Array1D<double> &y, const Array1D<double> &z, 
  bool datGeo ){

  int i, nDat = x.dim();
  double sse=0;
  double d2, YHat;
  if( dist_ == 0.0 ){
    MESSAGE "Dist_ = 0\n" ERROR;
  }
  vector<double>::iterator itX,
    itY,
    itHeight;

  for(i=0; i<nDat; i++){
    itY = yLoc_.begin();
    itHeight = atomHeight_.begin();
    YHat = 0;
    
    // The slow part of the following calculation may be the 
    // exponential function
    for( itX = xLoc_.begin(); itX < xLoc_.end(); itX++ ){
      
      d2 = SQR((*itX) - x[i]) + SQR((*itY) - y[i]);
      YHat += (*itHeight) * ( S_expm1( -0.5 * d2 / SQR(dist_) ) + 1 );
      itY++;  
      itHeight++;
    }
    if( ( z[i] <= 0.0 ) || ( YHat <= 0.0 ) ){
      MESSAGE "z[i]<=0 or Yhat <= 0\n" ERROR;
    }

    sse += SQR( log( z[i] / YHat ) );
  }
  return(sse);
}

/**************************************************************************
 **  Draw a new atom for the Levy process.  Draw the location (x,y)
 **  uniformly from the rectangular region, and draw the height z
 **  using the scale s.  Calculate the pdf (if ln = false) or the
 **  log-pdf (if ln = true).  Do not add the atom to the levy process.
 **/  
void LevyProc::drawPoint( 
  double *x, double *y, double *z,  // The location and height of the new atom
  double *pdf,     // The pdf
  double s,         // The scale for generating the weight of the new atom
  bool ln        // Whether to return the log pdf or the pdf
  ) {

  drawUnifLoc( x, y );
  (*z)   = epsilon_ + S_rexp() * s;     // New mass  z ~ eps+Exp(1/sca)
  (*pdf) = dBirth( (*x), (*y), (*z), s, ln);
}

/**************************************************************************
 ** Calculate the density of the atom birth proposal distribution
 **/  
double LevyProc::dBirth( 
  double x, double y, double z,  // The location and height of the atom
  double s,         // The scale with which the weight of the atom would be 
                  // generated
  bool ln        // Whether to return the log pdf or the pdf
  ) {

  double pdf = ( epsilon_-z ) / s - log( s * area );
  if(!ln) (pdf) = exp(pdf);
  return( pdf );
}

/**************************************************************************
 **  Add an atom at location (x,y) and with height z to the Levy
 **  process so that the atom has index ind in the new list of atoms.
 **  ind should be between -1 and length() inclusive; a value of -1
 **  is taken to be equal to length().  Return the index of the new atom.
 **/  
int LevyProc::addPoint( 
  double x, double y, double z,  // The location and height of the new atom
  int ind
  ) {

  if( (ind < -1) || (ind > nAtoms_) ){
    MESSAGE "Invalid atom index for insertion\n" ERROR;
  }
  if( z < epsilon_ ){
    MESSAGE "Cannot add a jump with height < epsilon" ERROR;
  }

  vector<double>::iterator iter = xLoc_.begin() + ind;
  if( (ind < nAtoms_) && (ind >= 0) ){
    xLoc_.insert( iter, x );  
    iter = yLoc_.begin() + ind;
    yLoc_.insert( iter, y );
    iter = atomHeight_.begin() + ind;
    atomHeight_.insert( iter, z );
  } else {
    // vector insert does not work for inserting at the end of the vector;
    // use push_back for this special case.
    xLoc_.push_back( x );
    yLoc_.push_back( y );
    atomHeight_.push_back( z );
  }
  nAtoms_ += 1;

  if (ind == -1)
    return( nAtoms_-1 );
  else
    return( ind );
}

/**************************************************************************
 **  Get the location and magnitude of the atom with index ind.
 **/  
void LevyProc::getPoint( 
  int ind, double *x, double *y, double *z
  ) {
  if( (ind < 0) || (ind >= nAtoms_) ){
    MESSAGE "Invalid atom indicator\n" ERROR;
  }
  (*x) = xLoc_[ind];
  (*y) = yLoc_[ind];
  (*z) = atomHeight_[ind];
}


/**************************************************************************
 **  Change the atom with index ind to location (x,y) and height z
 **/  
void LevyProc::changePoint( 
  int ind, double x, double y, double z
  ) {

  if( (ind < 0) || (ind >= nAtoms_) )
    MESSAGE "Invalid atom indicator\n" ERROR;
  if( z < epsilon_ )
    MESSAGE "cannot change point to have magnitude < epsilon\n" ERROR;

  // If necessary, reflect the atom location to be inside the limits
  while( (x < xlim_[0]) || (x > xlim_[1]) ) {
    if( x < xlim_[0] ) 
      x = 2 * xlim_[0] - x;   
    if( x > xlim_[1] ) 
      x = 2 * xlim_[1] - x;
  }
  while( (y < ylim_[0]) || (y > ylim_[1]) ) {
    if( y < ylim_[0] ) 
      y = 2 * ylim_[0] - y;   
    if( y > ylim_[1] ) 
      y = 2 * ylim_[1] - y;
  }

  // change the atom location and magnitude
  xLoc_[ind] = x;
  yLoc_[ind] = y;
  atomHeight_[ind] = z;
}

/**************************************************************************
 **  Delete the atom with index ind.
 **/  
void LevyProc::deletePoint( 
  int ind
  ) {
  if( (ind < 0) || (ind >= nAtoms_) ){
    MESSAGE "Invalid atom indicator for deletion\n" ERROR;
  }
  vector<double>::iterator iter = xLoc_.begin() + ind;
  xLoc_.erase( iter );
  iter = yLoc_.begin() + ind;
  yLoc_.erase( iter );
  iter = atomHeight_.begin() + ind;
  atomHeight_.erase( iter );

  nAtoms_ -= 1;
}


/**************************************************************************
 **  Write the Levy process atom locations and heights out to a
 **  file, appending to the end of the file.
 **/  
void LevyProc::appendToFile( FILE *outfile ){
  double x[1], y[1], height[1];
  for (int i=0; i<nAtoms_; i++){
    (*x) = xLoc_[i];
    fwrite( x, sizeof(double), 1, outfile );
    (*y) = yLoc_[i];
    fwrite( y, sizeof(double), 1, outfile );
    (*height) = atomHeight_[i];
    fwrite( height, sizeof(double), 1, outfile );
    //fprintf( outfile, "%10.4f %10.4f %15.4f\n", xLoc_[i], yLoc_[i], 
    //     atomHeight_[i] );
  }
}

/**************************************************************************
 **  Print the Levy process atom locations and heights to stdout
 **/  
void LevyProc::print(){
  for (int i=0; i<nAtoms_; i++){
    printf( "%f       %f       %f\n", xLoc_[i], yLoc_[i], 
	     atomHeight_[i] );
  }
}
