/*
*
* Template Numerical Toolkit (TNT)
*
* Mathematical and Computational Sciences Division
* National Institute of Technology,
* Gaithersburg, MD USA
*
*
* This software was developed at the National Institute of Standards and
* Technology (NIST) by employees of the Federal Government in the course
* of their official duties. Pursuant to title 17 Section 105 of the
* United States Code, this software is not subject to copyright protection
* and is in the public domain. NIST assumes no responsibility whatsoever for
* its use by other parties, and makes no guarantees, expressed or implied,
* about its quality, reliability, or any other characteristic.
*
*/


#include "moreArray1dUtils.h"
#include <cassert>
#include <cmath>
#include <fstream>
#include <vector>
#include <set>

using TNT::Array1D;
using std::cerr;
using std::endl;
using std::vector;
using std::set;


/* Divide a scalar by each element of an array. */
Array1D<double> TNT::operator/(double b, const Array1D<double> &A)
{
	int m = A.dim1();

	Array1D<double> C(m);

	for (int i=0; i<m; i++)
	  C[i] = b / A[i];

	return C;
 }	

/* Divide each element of an array by a scalar. */
Array1D<double> TNT::operator/(const Array1D<double> &A, double b){
  int m = A.dim1();
  
  Array1D<double> C(m);

  for (int i=0; i<m; i++)
    C[i] =  A[i] / b;
  
  return C;
}

  /* Exponentiation of an array to a power. */
Array1D<double> TNT::operator^(const Array1D<double> &A, double b){
	int m = A.dim1();

	Array1D<double> C(m);

	for (int i=0; i<m; i++)
	  C[i] = pow(A[i],b);

	return C;
}

/* Add a constant to an array */
Array1D<double> TNT::operator+(const Array1D<double> &A, double b)
{
	int m = A.dim1();

	Array1D<double> C(m);

	for (int i=0; i<m; i++)
	  C[i] = A[i] + b;

	return C;
 }	

/* Subtraction of arrays and constants */
Array1D<double> TNT::operator-(const Array1D<double> &A, double b){
	int m = A.dim1();

	Array1D<double> C(m);

	for (int i=0; i<m; i++)
	  C[i] = A[i] - b;

	return C;
}

/* Subtraction of arrays and constants */
Array1D<double> TNT::operator-(double a, const Array1D<double> &B){
	int m = B.dim1();

	Array1D<double> C(m);

	for (int i=0; i<m; i++)
	  C[i] = a - B[i];

	return C;
}

/* Comparison of arrays and constants */
Array1D<int> TNT::operator==(const Array1D<int> &A, int b){
	int m = A.dim1();

	Array1D<int> C(m);

	for (int i=0; i<m; i++)
	  C[i] = (A[i] == b);

	return C;
}

/* exponentiate each element of A */
Array1D<double> TNT::exp(const Array1D<double> &A){

  int m = A.dim1();
  
  Array1D<double> C(m);

  for (int i=0; i<m; i++)
    C[i] = std::exp(A[i]);

  return C;
  
}

/* Whether two scalars are approximately equal.  This is defined to
   mean that they are the same to within percentDiffAllowed.  The
   percent difference is found using the ratio of val1 to val2 and
   comparing to 0.  Therefore it matters which value is listed first,
   although for small epsilon it is not likely to change the result.
   If val2 is zero then it is tested whether val2 is within
   percentDiffAllowed of 0. */
bool TNT::approxEqual (double val1, double val2,
		  double percentDiffAllowed){
    if (val2 == 0) {
      if ( std::fabs( val1 - 0 ) > percentDiffAllowed)
	return false;
    } else {
      double percentDiff = (val1 / val2) - 1;
      if ( std::fabs( percentDiff ) > percentDiffAllowed)
	return false;
    }
    return true;
}

/* Tests whether two arrays are approximately equal.  Approximate
   equality is defined to mean that each element of A is approximately
   equal to the corresponding element of B, as evaluated by
   approxEqual(double, double, double) */
bool TNT::approxEqual (const Array1D<double> &A, const Array1D<double> &B, 
		  double percentDiffAllowed){
  int  n1 = A.dim1(), n2 = B.dim1();

  if (n1 != n2) 
    return false;

  for (int i=0; i<n1; i++)
    if ( !approxEqual (A[i], B[i], percentDiffAllowed) )
      return false;

  return true;
}

/* Tests whether every element of an array is approximately equal to
   a scalar.  Approximate equality is as evaluated by
   approxEqual(double, double, double) */
bool TNT::approxEqual (const Array1D<double> &A, double value,
		       double percentDiffAllowed){
  int n = A.dim();

  for (int i=0; i<n; i++)
    if ( !approxEqual (A[i], value, percentDiffAllowed))
      return false;

  return true;
}

/* concatenate two arrays */
Array1D<double>  TNT::concat(const Array1D<double> &A, const Array1D<double> &B)
  {
    int n1 = A.dim1(), n2 = B.dim1();

    Array1D<double> C (n1+n2);

    for (int i=0; i<n1; i++)
      C[i] = A[i];

    for (int i=0; i<n2; i++)
      C[i+n1] = B[i];

    return C;
  }



/* get the sum of an array.  If the array is null, return 0. */
double  TNT::sum(const Array1D<double> &A, int start, int end)
  {
    int n = A.dim1();
    if (n == 0) return 0;
    if (end == -1){
      end = A.dim1()-1;
    }    

    if ((start < 0) || (end >= n))
      cerr << "Sums of arrays must be from start index >= 0 to end index < length" <<endl;

    double s=0;
    for (int i=start; i<=end; i++)
      s += A[i];

    return s;
  }



/* get the sum of an array.  If the array is null, return 0. */
int  TNT::sum(const Array1D<int> &A, int start, int end)
  {
    int n = A.dim1();
    if (n == 0) return 0;
    if (end == -1){
      end = A.dim1()-1;
    }

    if ((start < 0) || (end >= n))
      cerr << "Sums of arrays must be from start index >= 0 to end index < length" <<endl;

    int s=0;
    for (int i=start; i<=end; i++)
      s += A[i];

    return s;
    }

/* get the product of an array.  If the array is null, return 1. */
double  TNT::prod(const Array1D<double> &A)
  {
    int n = A.dim1();
    if (n == 0) return 1;
    
    double s=1;

    for (int i=0; i<n; i++)
      s = s*A[i];

    return s;
  }


/* get the max value in an array, assuming that the array is made up
   of positive elements.  If the array is null, return 0. */
double  TNT::max(const Array1D<double> &A, int &maxValueIndex)
  {
    double maximum = 0;
    maxValueIndex = -1;

    int n = A.dim1();
    if (n == 0) return 0;

    for (int i=0; i<n; i++)
      if (A[i] > maximum){
	maximum = A[i];
	maxValueIndex = i;
      }

    return maximum;
  }

/* get the L2 norm of an array.  If the array is null, return 0. */
double  TNT::normL2(const Array1D<double> &A)
  {
    int n = A.dim1();
    if (n==0) return -1;

    double sum=0;

    for (int i=0; i<n; i++)
      sum += std::pow(A[i],2);

    return std::pow(sum,0.5);
  }


   
// Sort the array in descending order.  Sort in place.
void TNT::sort(Array1D<double> &A, bool descending){

  int i, j, array_size= A.dim();
  double temp;

  for (i = (array_size - 1); i >= 0; i--)
  {
    for (j = 1; j <= i; j++)
    {
      if (A[j-1] < A[j])
      {
        temp = A[j-1];
        A[j-1] = A[j];
        A[j] = temp;
      }
    }
  }
  if (!descending){
    Array1D<double> Atemp = A.copy();
    for (i=0; i<array_size; i++)
      A[i] = Atemp[array_size-i-1];
  }

}
   
/* The index of the first occurrence of b in A, up to a difference of
   0.00001.  If not found, returns -1.
*/
int TNT::find(const Array1D<double> &A, double b, double diff){
   int n = A.dim1();
   
   for (int i=0; i<n; i++){	
     if (approxEqual(A[i], b, diff))  return i;

   }
   return -1;
 }

/* The index of the first occurrence of b in A.  If not found, returns
   -1.
*/
int TNT::find(const Array1D<int> &A, int b){
   int n = A.dim1();
   
   for (int i=0; i<n; i++){	
     if (A[i] == b)  return i;

   }
   return -1;
 }

 
/* the vector of length numValues, with elements starting at start and with last element
   equal to end, equally spaced. */
TNT::Array1D<double> TNT::range(double start, double end, int numValues){
  int i;
  double increment;
  TNT::Array1D<double> increasingRange (numValues);

  if (start > end) {
    std::cerr << "Start should be less than end in range function";
    exit(1);
  }
  if (start == end){
    increasingRange[0] = start;
  }
  else{
    if (numValues <= 1){
      std::cerr << "The number of values should be at least two in the range function" << std::endl;
      exit(1);
    }
    else{
      increment = (end - start) / (numValues-1);
      for (i=0; i<numValues; i++)
	increasingRange[i] = i * increment + start;
    }
  }
  return increasingRange;

}




/* Create a vector of values, given a minimum, maximum, and the number
   of values desired.  Space the values equally on the log scale. */
Array1D<double> TNT::rangeInverseScale (double minimum, double maximum, 
					int numValues){

  // Cut on the 1/value scale
  double minInverseValue = 1 / maximum;
  double maxInverseValue = 1 / minimum;

  Array1D<double> inverseValues = 
    TNT::range(minInverseValue, maxInverseValue, numValues);

  Array1D<double> values = 1 / inverseValues;

  reverse(values);

  return values;

}

/* Create a vector of values, given a minimum, maximum, and the number
   of values desired.  Space the values equally on the log scale. */
Array1D<double> TNT::rangeLogScale (double minimum, double maximum, 
				    int numValues){

  // Cut on the log scale
  double minLog = log( minimum );
  double maxLog = log( maximum );

  Array1D<double> logValues = TNT::range(minLog, 
					 maxLog, numValues);

  Array1D<double> values = exp(logValues);
  
  return values;

}



/* Create a vector of values, given a minimum, maximum, and the number
   of values desired.  Space the values equally on the log 1/value
   scale. */
Array1D<double> TNT::rangeInverseLogScale 
(double minimum, double maximum, 
 int numValues){

  // Cut on the 1/value scale
  double minInverseValue = 1 / maximum;
  double maxInverseValue = 1 / minimum;

  Array1D<double> inverseValues = 
    TNT::rangeLogScale(minInverseValue, maxInverseValue, numValues);

  Array1D<double> values = 1 / inverseValues;

  reverse(values);

  return values;

}



// write as a R/S-plus style matrix
void TNT::writeArray(const Array1D <double> &array, std::string filename)
{
  std::ofstream dataStream (filename.c_str());
  if (dataStream.fail())
    {
      std::cerr << "Unable to open output file " << filename << std::endl;
      exit(1);
    }
  writeArray (array, dataStream);  
  dataStream.close();
}


// write as a R/S-plus style matrix
void TNT::writeArray(const Array1D <double> &array, std::ostream &out)
{
  int i, nElements = array.dim();

  // write out the data on one line
  for (i = 0; i < nElements; i++){

    out << array[i];
    // if not at end of file, put a newline.
    if (i < nElements-1)   out << std::endl; 
  }
  
}

// read from a file as a R/S-plus style matrix
Array1D<double> TNT::readDoubleArray(std::string filename)
{
  std::ifstream dataStream (filename.c_str());
  if (dataStream.fail())
    {
      std::cerr << "Unable to open input file " << filename << std::endl;
      exit(1);
    }
  Array1D<double> arr = readDoubleArray (dataStream);  
  dataStream.close();

  return(arr);
}

// read from a file as a R/S-plus style matrix
Array1D<double> TNT::readDoubleArray(std::istream &in)
{
  std::vector<double> *arrVec = new std::vector<double>;
  double temp;

  while(in >> temp){
    arrVec->push_back(temp);
  }

  int arrLen = arrVec->size();
  Array1D<double> arr (arrLen);
  for (int i = 0; i < arrLen; i++){
    arr[i] = arrVec->at(i);
  }

  return(arr);
}

// read from a file as a R/S-plus style matrix
Array1D<int> TNT::readIntArray(std::string filename)
{
  std::ifstream dataStream (filename.c_str());
  if (dataStream.fail())
    {
      std::cerr << "Unable to open input file " << filename << std::endl;
      exit(1);
    }
  Array1D<int> arr = readIntArray (dataStream);  
  dataStream.close();

  return(arr);
}

// read from a file as a R/S-plus style matrix
Array1D<int> TNT::readIntArray(std::istream &in)
{
  std::vector<int> *arrVec = new std::vector<int>;
  int temp;

  while(in >> temp){
    arrVec->push_back(temp);
  }

  int arrLen = arrVec->size();
  Array1D<int> arr (arrLen);
  for (int i = 0; i < arrLen; i++){
    arr[i] = arrVec->at(i);
  }

  return(arr);
}


// Reverse the order of the elements
void TNT::reverse(Array1D <double> &array){

  int nElements = array.dim();
  Array1D<double> temp = array.copy();
  for (int i=0; i<nElements; i++)
    array[i] = temp[nElements - 1 - i];
   
}



